@@ -1,49 +1,30 @@
-""" A simple genetic algorithm for parameter search """
-
+"""A simple genetic algorithm for parameter search."""
 import random
+
 import numpy as np
 
-from kernel_tuner.strategies.minimize import _cost_func
-from kernel_tuner.searchspace import Searchspace
 from kernel_tuner import util
+from kernel_tuner.searchspace import Searchspace
+from kernel_tuner.strategies import common
+from kernel_tuner.strategies.common import CostFunc
+
+_options = dict(
+    popsize=("population size", 20),
+    maxiter=("maximum number of generations", 100),
+    method=("crossover method to use, choose any from single_point, two_point, uniform, disruptive_uniform", "uniform"),
+    mutation_chance=("chance to mutate is 1 in mutation_chance", 10),
+)
 
 
-def tune(runner, kernel_options, device_options, tuning_options):
-    """ Find the best performing kernel configuration in the parameter space
-
-    :params runner: A runner from kernel_tuner.runners
-    :type runner: kernel_tuner.runner
-
-    :param kernel_options: A dictionary with all options for the kernel.
-    :type kernel_options: kernel_tuner.interface.Options
-
-    :param device_options: A dictionary with all options for the device
-        on which the kernel should be tuned.
-    :type device_options: kernel_tuner.interface.Options
-
-    :param tuning_options: A dictionary with all options regarding the tuning
-        process.
-    :type tuning_options: kernel_tuner.interface.Options
-
-    :returns: A list of dictionaries for executed kernel configurations and their
-        execution times. And a dictionary that contains information
-        about the hardware/software environment on which the tuning took place.
-    :rtype: list(dict()), dict()
-
-    """
+def tune(searchspace: Searchspace, runner, tuning_options):
 
     options = tuning_options.strategy_options
-    pop_size = options.get("popsize", 20)
-    generations = options.get("maxiter", 50)
-    crossover = supported_methods[options.get("method", "uniform")]
-    mutation_chance = options.get("mutation_chance", 10)
-
-    tuning_options["scaling"] = False
+    pop_size, generations, method, mutation_chance = common.get_options(options, _options)
+    crossover = supported_methods[method]
 
     best_score = 1e20
-    results = []
+    cost_func = CostFunc(searchspace, tuning_options, runner)
 
-    searchspace = Searchspace(tuning_options, runner.dev.max_threads)
     population = list(list(p) for p in searchspace.get_random_sample(pop_size))
 
     for generation in range(generations):
@@ -52,11 +33,11 @@ def tune(runner, kernel_options, device_options, tuning_options):
         weighted_population = []
         for dna in population:
             try:
-                time = _cost_func(dna, kernel_options, tuning_options, runner, results, check_restrictions=False)
+                time = cost_func(dna, check_restrictions=False)
             except util.StopCriterionReached as e:
                 if tuning_options.verbose:
                     print(e)
-                return results, runner.dev.get_environment()
+                return cost_func.results
 
             weighted_population.append((dna, time))
 
@@ -64,8 +45,8 @@ def tune(runner, kernel_options, device_options, tuning_options):
         weighted_population.sort(key=lambda x: x[1])
 
         # 'best_score' is used only for printing
-        if tuning_options.verbose and results:
-            best_score = util.get_best_config(results, tuning_options.objective, tuning_options.objective_higher_is_better)[tuning_options.objective]
+        if tuning_options.verbose and cost_func.results:
+            best_score = util.get_best_config(cost_func.results, tuning_options.objective, tuning_options.objective_higher_is_better)[tuning_options.objective]
 
         if tuning_options.verbose:
             print("Generation %d, best_score %f" % (generation, best_score))
@@ -89,11 +70,14 @@ def tune(runner, kernel_options, device_options, tuning_options):
 
         # could combine old + new generation here and do a selection
 
-    return results, runner.dev.get_environment()
+    return cost_func.results
+
+
+tune.__doc__ = common.get_strategy_docstring("Genetic Algorithm", _options)
 
 
 def weighted_choice(population, n):
-    """Randomly select n unique individuals from a weighted population, fitness determines probability of being selected"""
+    """Randomly select n unique individuals from a weighted population, fitness determines probability of being selected."""
 
     def random_index_betavariate(pop_size):
         # has a higher probability of returning index of item at the head of the list
@@ -102,7 +86,7 @@ def weighted_choice(population, n):
         return int(random.betavariate(alpha, beta) * pop_size)
 
     def random_index_weighted(pop_size):
-        """use weights to increase probability of selection"""
+        """Use weights to increase probability of selection."""
         weights = [w for _, w in population]
         # invert because lower is better
         inverted_weights = [1.0 / w for w in weights]
@@ -124,26 +108,28 @@ def weighted_choice(population, n):
     return [population[ind][0] for ind in chosen]
 
 
-def mutate(dna, mutation_chance, searchspace: Searchspace):
-    """Mutate DNA with 1/mutation_chance chance"""
-
+def mutate(dna, mutation_chance, searchspace: Searchspace, cache=True):
+    """Mutate DNA with 1/mutation_chance chance."""
     # this is actually a neighbors problem with Hamming distance, choose randomly from returned searchspace list
     if int(random.random() * mutation_chance) == 0:
-        neighbors = searchspace.get_neighbors(tuple(dna), neighbor_method='Hamming')
+        if cache:
+            neighbors = searchspace.get_neighbors(tuple(dna), neighbor_method="Hamming")
+        else:
+            neighbors = searchspace.get_neighbors_no_cache(tuple(dna), neighbor_method="Hamming")
         if len(neighbors) > 0:
             return list(random.choice(neighbors))
     return dna
 
 
 def single_point_crossover(dna1, dna2):
-    """crossover dna1 and dna2 at a random index"""
+    """Crossover dna1 and dna2 at a random index."""
     # check if you can do the crossovers using the neighbor index: check which valid parameter configuration is closest to the crossover, probably best to use "adjacent" as it is least strict?
     pos = int(random.random() * (len(dna1)))
     return (dna1[:pos] + dna2[pos:], dna2[:pos] + dna1[pos:])
 
 
 def two_point_crossover(dna1, dna2):
-    """crossover dna1 and dna2 at 2 random indices"""
+    """Crossover dna1 and dna2 at 2 random indices."""
     if len(dna1) < 5:
         start, end = 0, len(dna1)
     else:
@@ -155,7 +141,7 @@ def two_point_crossover(dna1, dna2):
 
 
 def uniform_crossover(dna1, dna2):
-    """randomly crossover genes between dna1 and dna2"""
+    """Randomly crossover genes between dna1 and dna2."""
     ind = np.random.random(len(dna1)) > 0.5
     child1 = [dna1[i] if ind[i] else dna2[i] for i in range(len(ind))]
     child2 = [dna2[i] if ind[i] else dna1[i] for i in range(len(ind))]
@@ -163,7 +149,7 @@ def uniform_crossover(dna1, dna2):
 
 
 def disruptive_uniform_crossover(dna1, dna2):
-    """disruptive uniform crossover
+    """Disruptive uniform crossover.
 
     uniformly crossover genes between dna1 and dna2,
     with children guaranteed to be different from parents,
@@ -189,5 +175,5 @@ supported_methods = {
     "single_point": single_point_crossover,
     "two_point": two_point_crossover,
     "uniform": uniform_crossover,
-    "disruptive_uniform": disruptive_uniform_crossover
+    "disruptive_uniform": disruptive_uniform_crossover,
 }
