@@ -1,16 +1,14 @@
-from collections import OrderedDict
-
 from kernel_tuner.hyper import tune_hyper_params
 
-from .test_runners import env, cache_filename
+from .test_runners import cache_filename, env  # noqa: F401
 
 
 def test_hyper(env):
 
-    hyper_params = OrderedDict()
+    hyper_params = dict()
     hyper_params["popsize"] = [5]
     hyper_params["maxiter"] = [5, 10]
-    hyper_params["crossover"] = ["uniform"]
+    hyper_params["method"] = ["uniform"]
     hyper_params["mutation_chance"] = [10]
 
     target_strategy = "genetic_algorithm"
